from collections import Counter
from functools import cached_property, cache
from itertools import product, permutations
import numpy as np
from typing import Iterable, Optional
from .circuitree import CircuiTree

__all__ = ["SimpleNetworkTree", "DimerNetworkTree"]

"""Classes for modeling different types of circuits."""


class SimpleNetworkGrammar:
    """A class implementing the grammar for simple network circuits. Can be combined with
    a CircuiTree or MultithreadedCircuiTree class that implements the get_reward() method
    ."""

    def __init__(
        self,
        components: Iterable[Iterable[str]],
        interactions: Iterable[str],
        max_interactions: Optional[int] = None,
        **kwargs,
    ):
        if len(set(c[0] for c in components)) < len(components):
            raise ValueError("First character of each component must be unique")
        if len(set(c[0] for c in interactions)) < len(interactions):
            raise ValueError("First character of each interaction must be unique")

        super().__init__(**kwargs)

        self.components = components
        self.component_map = {c[0]: c for c in self.components}
        self.interactions = interactions
        self.interaction_map = {ixn[0]: ixn for ixn in self.interactions}

        if max_interactions is None:
            self.max_interactions = len(self.components) ** 2  # all possible edges
        else:
            self.max_interactions = max_interactions

        self._non_serializable_attrs.extend(
            ["component_map", "interaction_map", "edge_options", "_recolor"]
        )

    @cached_property
    def edge_options(self):
        return [
            [c1[0] + c2[0] + ixn[0] for ixn in self.interactions]
            for c1, c2 in product(self.components, self.components)
        ]

    def get_actions(self, genotype: str) -> Iterable[str]:
        if self.is_terminal(genotype):
            return list()

        # Terminating assembly is always an option
        actions = ["*terminate*"]

        components, interactions_joined = genotype.strip("*").split("::")
        interactions = set(ixn[:2] for ixn in interactions_joined.split("_"))
        for component in self.components:
            c0 = component[0]
            if c0 not in components:
                actions.append(c0)
        for action_group in self.edge_options:
            if action_group:
                c1, c2, _ = action_group[0]
                if (
                    (c1 in components)
                    and (c2 in components)
                    and (c1 + c2) not in interactions
                ):
                    for action in action_group:
                        actions.append(action)

        return actions

    def do_action(self, genotype: str, action: str) -> str:
        if action == "*terminate*":
            new_genotype = "*" + genotype
        else:
            # Root node
            if genotype == ".":
                components = list()
                interactions = list()
            else:
                components, interactions = genotype.split("::")
            if len(action) == 1:
                new_genotype = "".join([components, action, "::", interactions])
            elif len(action) == 3:
                delim = ("", "_")[bool(interactions)]
                new_genotype = "::".join(
                    [components, delim.join([interactions, action])]
                )
        return new_genotype

    @staticmethod
    def is_terminal(genotype: str) -> bool:
        return genotype.startswith("*")

    @staticmethod
    def get_unique_state(genotype: str) -> str:
        prefix = "*" if genotype.startswith("*") else ""
        components, interactions = genotype.strip("*").split("::")
        unique_components = "".join(sorted(components))
        unique_interactions = "_".join(sorted(interactions.split("_")))
        return prefix + unique_components + "::" + unique_interactions

    @staticmethod
    def parse_genotype(genotype: str, nonterminal_ok: bool = False):
        if not genotype.startswith("*") and not nonterminal_ok:
            raise ValueError(
                f"Assembly incomplete. Genotype {genotype} is not a terminal genotype."
            )
        components, interaction_codes = genotype.strip("*").split("::")
        component_indices = {c: i for i, c in enumerate(components)}

        interactions = [i for i in interaction_codes.split("_") if i]

        activations = []
        inhbitions = []
        for left, right, ixn in interactions:
            if ixn.lower() == "a":
                activations.append((component_indices[left], component_indices[right]))
            elif ixn.lower() == "i":
                inhbitions.append((component_indices[left], component_indices[right]))

        activations = np.array(activations, dtype=np.int_)
        inhbitions = np.array(inhbitions, dtype=np.int_)

        return components, activations, inhbitions


class SimpleNetworkTree(SimpleNetworkGrammar, CircuiTree):
    """
    SimpleNetworkTree
    =================
    Models a circuit as a set of components and strictly pairwise interactions between them.
    The circuit topology (also referred to as the "state" during search) is encoded using a
    string representation with the following rules:
        - Components are represented by a single uppercase character
        - Interactions are represented by a three-character string
            - The first two characters are the components involved. Order is assumed to
                matter.
            - The third character (lowercase) is the type of interaction. A common use case
                would have "a" for activation and "i" for inhibition.
        - Multiple interactions are separated by underscores ``_``
        - Components and interactions are separated by double colons ``::``
        - A terminal assembly is denoted with a leading asterisk ``*``

        For example, the following string represents a terminally assembled circuit
        that encodes an incoherent feed-forward loop:
            ``*ABC::ABa_ACa_BCi``
    """


class DimersMixin:
    """A mixin class for dimerizing network circuits. Can be combined with the CircuiTree
    or MultithreadedCircuiTree base classes."""

    @cached_property
    def dimer_options(self):
        dimers = set()
        monomers = self.components + self.regulators
        for c1, c2 in product(monomers, monomers):
            if c1 in self.regulators and c2 in self.regulators:
                continue
            dimers.add("".join(sorted([c1[0], c2[0]])))
        return list(dimers)

    @cached_property
    def edge_options(self):
        return [
            [d + ixn[0] + c[0] for ixn in self.interactions]
            for d, c in product(self.dimer_options, self.components)
        ]

    @staticmethod
    def is_terminal(genotype: str) -> bool:
        return genotype.startswith("*")

    def get_actions(self, genotype: str) -> Iterable[str]:
        if self.is_terminal(genotype):
            return list()

        # Terminating assembly is always an option
        actions = ["*terminate*"]

        components, regulators, interactions_joined = self.get_genotype_parts(genotype)
        actions.extend(list(set(self.components) - set(components)))
        regulators_to_add = set(self.regulators) - set(regulators)
        actions.extend([f"+{r}" for r in regulators_to_add])
        n_bound = Counter(ixn[3] for ixn in interactions_joined.split("_") if ixn)
        for action_group in self.edge_options:
            if action_group:
                promoter = action_group[0][-1]
                if n_bound[promoter] < self.n_binding_sites:
                    actions.extend(action_group)

        return actions

    def do_action(self, genotype: str, action: str) -> str:
        action_len = len(action)
        components, regulators, interactions = self.get_genotype_parts(genotype)
        if action == "*terminate*":
            components = "*" + components
        elif action_len == 1:
            components = components + action
        elif action_len == 2 and action[0] == "+":
            regulators = regulators + action[1]
        elif action_len == 4:
            ixn_list = interactions.split("_") if interactions else []
            interactions = "_".join(ixn_list + [action])
        return components + "+" + regulators + "::" + interactions

    @staticmethod
    @cache
    def get_unique_state(genotype: str) -> str:
        components_and_regulators, interactions = genotype.split("::")
        components, regulators = components_and_regulators.split("+")
        if components.startswith("*"):
            prefix = "*"
            components = components[1:]
        else:
            prefix = ""
        unique_components = "".join(sorted(components))
        unique_regulators = "".join(sorted(regulators))
        unique_interaction_list = []
        for ixn in interactions.split("_"):
            *monomers, logic, promoter = ixn
            unique_ixn = "".join(sorted(monomers)) + logic + promoter
            unique_interaction_list.append(unique_ixn)
        unique_interactions = "_".join(sorted(unique_interaction_list))
        unique_genotype = "".join(
            [
                prefix,
                unique_components,
                "+",
                unique_regulators,
                "::",
                unique_interactions,
            ]
        )
        return unique_genotype

    @staticmethod
    def get_genotype_parts(genotype: str):
        components_and_regulators, interactions = genotype.strip("*").split("::")
        components, regulators = components_and_regulators.split("+")
        return components, regulators, interactions

    @staticmethod
    def parse_genotype(genotype: str, nonterminal_ok: bool = False):
        if not genotype.startswith("*") and not nonterminal_ok:
            raise ValueError(
                f"Assembly incomplete. Genotype {genotype} is not a terminal genotype."
            )

        components_and_regulators, interaction_codes = genotype.strip("*").split("::")
        components, regulators = components_and_regulators.split("+")
        component_indices = {c: i for i, c in enumerate(components + regulators)}

        interactions = interaction_codes.split("_") if interaction_codes else []
        activations = []
        inhbitions = []
        for *monomers, regulation_type, promoter in interactions:
            m1, m2 = sorted(monomers)
            ixn_tuple = (
                component_indices[m1],
                component_indices[m2],
                component_indices[promoter],
            )
            if regulation_type.lower() == "a":
                activations.append(ixn_tuple)
            elif regulation_type.lower() == "i":
                inhbitions.append(ixn_tuple)
            else:
                raise ValueError(
                    f"Unknown regulation type {regulation_type} in {genotype}"
                )
        activations = np.array(activations, dtype=np.int_)
        inhbitions = np.array(inhbitions, dtype=np.int_)
        return components, regulators, activations, inhbitions

    @cached_property
    def _recolor_components(self):
        return [dict(zip(self.components, p)) for p in permutations(self.components)]

    @cached_property
    def _recolor_regulators(self):
        return [dict(zip(self.regulators, p)) for p in permutations(self.regulators)]

    @cached_property
    def _recolorings(self):
        return [
            rc | rr
            for rc, rr in product(self._recolor_components, self._recolor_regulators)
        ]

    @staticmethod
    def _recolor(mapping, code):
        return "".join([mapping.get(char, char) for char in code])

    def get_interaction_recolorings(self, genotype: str) -> list[str]:
        *_, interactions = self.get_genotype_parts(genotype)
        interaction_recolorings = (
            "_".join(
                sorted([self._recolor(mapping, ixn) for ixn in interactions.split("_")])
            ).strip("_")
            for mapping in self._recolorings
        )
        return interaction_recolorings

    def get_component_recolorings(self, genotype: str) -> list[str]:
        components, *_ = self.get_genotype_parts(genotype)
        component_recolorings = (
            "".join(sorted(self._recolor(mapping, components)))
            for mapping in self._recolor_components
        )
        return component_recolorings

    def get_regulator_recolorings(self, genotype: str) -> list[str]:
        _, regulators, *_ = self.get_genotype_parts(genotype)
        regulator_recolorings = (
            "".join(sorted(self._recolor(mapping, regulators)))
            for mapping in self._recolor_components
        )
        return regulator_recolorings

    def get_recolorings(self, genotype: str) -> Iterable[str]:
        prefix = "*" if self.is_terminal(genotype) else ""

        _, regulators, *_ = self.get_genotype_parts(genotype)
        if regulators:
            return (
                f"{prefix}{c}+{r}::{i}"
                for c, r, i in zip(
                    self.get_component_recolorings(genotype),
                    self.get_regulator_recolorings(genotype),
                    self.get_interaction_recolorings(genotype),
                )
            )
        else:
            return (
                f"{prefix}{c}::{i}"
                for c, i in zip(
                    self.get_component_recolorings(genotype),
                    self.get_interaction_recolorings(genotype),
                )
            )

    @cache
    def get_unique_state(self, genotype: str) -> str:
        return min(self.get_recolorings(genotype))

    @cache
    def _motif_recolorings(self, motif: str) -> list[set[str]]:
        if ("+" in motif) or ("::" in motif) or ("*" in motif):
            raise ValueError(
                "Motif code should only contain interactions, no components or "
                "regulators"
            )

        return [
            set(recoloring.split("_"))
            for recoloring in self.get_interaction_recolorings(motif)
        ]

    def has_motif(self, state, motif):
        if ("::" in motif) or ("*" in motif):
            raise ValueError(
                "Motif code should only contain interactions, no components"
            )
        if "::" not in state:
            raise ValueError(
                "State code should contain both components and interactions"
            )

        interaction_code = state.split("::")[1]
        if not interaction_code:
            return False
        state_interactions = set(interaction_code.split("_"))

        for motif_interactions_set in self._motif_recolorings(motif):
            if motif_interactions_set.issubset(state_interactions):
                return True
        return False


class DimerNetworkTree(DimersMixin, CircuiTree):
    """
    DimerNetworkTree
    =================
    A CircuiTree for the design space of dimerizing TF networks. Intended to recapitulate
    the dimerization of zinc-finger proteins.

    Models a system of dimerizing transcription factors (e.g. zinc-fingers) that regulate
    each other's transcription. The circuit consists a set of ``components``, which
    represent transcription factors that are being regulated. Components can form homo-
    or heterodimers that bind to a component's promoter region and regulate
    transcription. There is also a set of ``regulators``, which can dimerize and regulate
    transcription but are not themselves regulated. Regulator-regulator homodimers and
    regulator-component heterodimers can act as TFs, but regulator-regulator homodimers
    are assumed to be inactive.

    The circuit topology (also referred to as the "state" during search) is encoded using
    a string representation (aka "genotype") with the following rules:
        - Components and regulators are represented by single uppercase characters
        - Interactions are represented by a 4-character string
            - Characters 1-2 (uppercase): the dimerizing species (components/regulators)
            - Character 3 (lowercase): the type of regulation upon binding
            - Character 4 (uppercase): the target of regulation (a component)
        - Components are separated from regulators by a ``+``
        - Components/regulators are separated from interactions by a ``::``
        - Interactions are separated from one another by underscores ``_``
        - A terminal assembly is denoted with a leading asterisk ``*``

        For example, the following string represents a 2-component MultiFate system that
        has not been fully assembled (lacks the terminal asterisk):

            ``AB+::AAa_BBa``

        While the following string represents a terminally assembled 2-component
        MultiFate system with a regulator L that flips the system into the A state:

            ``*AB+L::AAa_ALa_BBa_BLi``

    """

    def __init__(
        self,
        components: Iterable[str],
        regulators: Iterable[str],
        interactions: Iterable[str],
        n_binding_sites: int = 2,
        **kwargs,
    ):
        if len(set(c[0] for c in components)) < len(components):
            raise ValueError("First character of each component must be unique")
        if len(set(c[0] for c in regulators)) < len(regulators):
            raise ValueError("First character of each component must be unique")
        if len(set(c[0] for c in interactions)) < len(interactions):
            raise ValueError("First character of each interaction must be unique")

        super().__init__(**kwargs)

        self.components = components
        self.component_map = {c[0]: c for c in self.components}
        self.regulators = regulators
        self.regulator_map = {r[0]: r for r in self.regulators}
        self.interactions = interactions
        self.interaction_map = {ixn[0]: ixn for ixn in self.interactions}

        self.n_binding_sites = n_binding_sites

        self._non_serializable_attrs.extend(
            ["component_map", "regulator_map", "interaction_map"]
        )



from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from dataclasses import dataclass, field
from itertools import chain
import json
from pathlib import Path
import networkx as nx
from numpy.random import default_rng, SeedSequence
import numpy as np
import pandas as pd
from multiprocessing import cpu_count
from typing import Any, Callable, Iterable, Optional, Sequence

from .modularity import tree_modularity, tree_modularity_estimate
from .circuitree import CircuiTree, accumulate_visits_and_rewards, ucb_score
from .utils import DefaultMapping

__all__ = [
    "MultithreadedCircuiTree",
    "search_mcts",
    "ParameterTable",
    "TranspositionTable",
    "ParallelTree",
]


class MultithreadedCircuiTree(ABC):
    def __init__(
        self,
        root: str,
        threads: Optional[int] = None,
        seed: int = 2023,
        exploration_constant: Optional[float] = None,
        graph: Optional[nx.DiGraph] = None,
        **kwargs,
    ):
        if exploration_constant is None:
            self.exploration_constant = np.sqrt(2)
        else:
            self.exploration_constant = exploration_constant

        if threads is None:
            threads = cpu_count()
        self.threads = threads

        self.seed = seed
        seq = SeedSequence(seed)
        self._random_generators = [default_rng(s) for s in seq.spawn(threads)]

        self.root = root
        if graph is None:
            self.graph = nx.DiGraph()
            self.graph.add_node(self.root, visits=0, reward=0)
        else:
            self.graph = graph
            if self.root not in self.graph:
                raise ValueError(
                    f"Supplied graph does not contain the root node: {root}"
                )
        self.graph.root = self.root

        # Attributes that should not be saved to file
        self._non_serializable_attrs = [
            "_non_serializable_attrs",
            "_random_generators",
            "graph",
        ]

    @property
    def default_attrs(self):
        return dict(visits=0, reward=0)

    def _do_action(self, state: Any, action: Any):
        new_state = self.do_action(state, action)
        return self.get_unique_state(new_state)

    @abstractmethod
    def get_actions(self, state: Any) -> Iterable[Any]:
        raise NotImplementedError

    @abstractmethod
    def do_action(self, state: Any, action: Any) -> Any:
        raise NotImplementedError

    @abstractmethod
    def is_terminal(self, state) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_reward(self, state) -> float | int:
        raise NotImplementedError

    @abstractmethod
    def get_unique_state(self, state: Any) -> Any:
        raise NotImplementedError

    def is_success(self, state) -> bool:
        """Must be defined to calculate modularity."""
        raise NotImplementedError

    def get_ucb_score(self, parent, child):
        return ucb_score(self.graph, parent, child, self.exploration_constant)

    def select(self, thread_idx: int):
        node = self.root
        selection_path = [node]
        while not self.is_leaf(node):
            node = self.best_child(thread_idx, node)
            selection_path.append(node)
        return selection_path

    def expand(self, thread_idx: int, selection_path: list[Any]) -> list[Any]:
        """Expands a candidate selected node and returns the resulting selected node and
        selection path.

        If the candidate is non-terminal, selects a random child and appends it to the
        selection path. Otherwise, does nothing."""
        node = selection_path[-1]
        actions = self.get_actions(node)
        if not actions:
            return selection_path
        children = [self._do_action(node, action) for action in actions]
        for action, child in zip(actions, children):
            if child not in self.graph.nodes:
                self.graph.add_node(child, **self.default_attrs)
                self.graph.add_edge(node, child, action=action, **self.default_attrs)

        rg = self._random_generators[thread_idx]
        selection_path.append(rg.choice(children))
        return selection_path

    def simulate(self, thread_idx, node, **kwargs) -> tuple[Any, float | int]:
        sim_node = node
        rg: np.random.Generator = self._random_generators[thread_idx]
        while not self.is_terminal(sim_node):
            action = rg.choice(self.get_actions(sim_node))
            sim_node = self._do_action(sim_node, action)
        reward = self.get_reward(sim_node, **kwargs)
        return sim_node, reward

    def backpropagate_reward(self, selection_path: list, reward: float | int):
        node = selection_path.pop()
        while selection_path:
            parent = selection_path.pop()
            self.graph.edges[parent, node]["reward"] += reward
            node = parent
        self.graph.nodes[node]["reward"] += reward  # root node

    def is_leaf(self, node):
        return self.graph.out_degree(node) == 0

    def best_child(self, thread_idx, node):
        rg: np.random.Generator = self._random_generators[thread_idx]
        children = list(self.graph.neighbors(node))
        rg.shuffle(children)
        scores = [self.get_ucb_score(node, child) for child in children]
        best = children[np.argmax(scores)]
        return best

    def backpropagate_visit(self, selection_path: list) -> None:
        for node in selection_path[::-1]:
            self.graph.nodes[node]["visits"] += 1

    def traverse(self, thread_idx: int, **kwargs):
        selection_path = self.select(thread_idx)
        selection_path = self.expand(thread_idx, selection_path)
        sel_node = selection_path[-1]
        self.backpropagate_visit(selection_path)
        sim_node, reward = self.simulate(sel_node, **kwargs)
        self.backpropagate_reward(selection_path.copy(), reward)
        return selection_path, reward, sim_node

    def accumulate_visits_and_rewards(self, graph: Optional[nx.DiGraph] = None):
        _accumulated = self.graph if graph is None else graph
        accumulate_visits_and_rewards(_accumulated)
        if graph is None:
            return _accumulated

    @property
    def terminal_states(self):
        return (node for node in self.graph.nodes if self.is_terminal(node))

    def bfs_iterator(self, root=None, shuffle=False):
        root = self.root if root is None else root
        return chain(*(l for l in nx.bfs_layers(self.graph, root)))

    def modularity(self) -> float:
        return tree_modularity(self.graph, self.root, self.is_terminal, self.is_success)

    def modularity_estimate(self) -> float:
        return tree_modularity_estimate(self.graph, self.root)

    def to_file(
        self,
        gml_file: str | Path,
        json_file: Optional[str | Path] = None,
        save_attrs: Optional[Iterable[str]] = None,
        **kwargs,
    ):
        gml_target = Path(gml_file).with_suffix(".gml")
        nx.write_gml(self.graph, gml_target, **kwargs)

        if json_file is not None:
            if save_attrs is None:
                keys = set(self.__dict__.keys()) - set(self._non_serializable_attrs)
            else:
                keys = set(save_attrs)
                if non_serializable := (keys & set(self._non_serializable_attrs)):
                    repr_non_ser = ", ".join(non_serializable)
                    raise ValueError(
                        f"Attempting to save non-serializable attributes: {repr_non_ser}."
                    )

            attrs = {k: v for k, v in self.__dict__.items() if k in keys}

            json_target = Path(json_file).with_suffix(".json")
            with json_target.open("w") as f:
                json.dump(attrs, f, indent=4)

            return gml_target, json_target

        else:
            return gml_target

    @classmethod
    def from_file(
        cls, graph_gml: str | Path, attrs_json: Optional[str | Path] = None, **kwargs
    ):
        if attrs_json is not None:
            with open(attrs_json, "r") as f:
                kwargs.update(json.load(f))

        graph = nx.read_gml(graph_gml)

        return cls(graph=graph, **kwargs)


def search_mcts(
    mtree: MultithreadedCircuiTree,
    thread_idx: int,
    n_steps: int,
    callback: Optional[Callable] = None,
    callback_every: int = 1,
    return_metrics: Optional[bool] = None,
    **kwargs,
):
    if callback is None:
        callback = lambda *a, **kw: None

    m0 = callback(mtree, 0, [None], None, None)
    if return_metrics is None:
        return_metrics = m0 is not None

    metrics = [m0]
    for iteration in range(1, n_steps + 1):
        selection_path, reward, sim_node = mtree.traverse(thread_idx, **kwargs)
        if iteration % callback_every == 0:
            m = callback(mtree, iteration, selection_path, reward, sim_node)
            if return_metrics:
                metrics.append(m)

    if return_metrics:
        return mtree, metrics
    else:
        return mtree


@dataclass
class ParameterTable:
    seeds: Iterable[int]
    initial_conditions: Optional[Sequence[Any]] = field(default_factory=list)
    parameter_sets: Optional[Sequence[Any]] = field(default_factory=list)
    wrap_index: bool = True

    """Stores a table of parameter sets that are used as inputs for a stochastic game.
    For each visit there is a random seed, a set of initial conditions, and a set of 
    parameter values."""

    def __post_init__(self):
        n = len(self.seeds)
        n_init = len(self.initial_conditions)
        n_params = len(self.parameter_sets)
        if n_init != n or n_params != n:
            raise ValueError(
                "Number of initial conditions, parameter sets, and random seeds "
                "must be equal."
            )

    def __len__(self):
        return len(self.seeds)

    def __getitem__(self, index):
        if self.wrap_index:
            index = index % len(self)
        return (
            self.seeds[index],
            self.initial_conditions[index],
            self.parameter_sets[index],
        )

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        seed_col: str = "seed",
        param_cols: Optional[Iterable[str]] = None,
        init_cols: Optional[Iterable[str]] = None,
    ):
        if df.index.name == seed_col:
            df.reset_index(inplace=True)
        elif seed_col not in df.columns:
            raise ValueError(
                f"Dataframe does not have a column or index named '{seed_col}'."
            )

        kw = {}
        if init_cols:
            kw["initial_conditions"] = list(df[init_cols].values)
        if param_cols:
            kw["parameter_sets"] = list(df[param_cols].values)

        return cls(df[seed_col].tolist(), **kw)


class TranspositionTable:
    """
    Stores the history of reward payouts at each visit to each state in a decision tree.
    Used to store results and simulate playouts of a stochastic game.

    In a stochastic game, each visit to a terminal state yields a different result.
    For a given random series of playouts, the reward values to a terminal state `s` are
    stored in a list `vals`. The reward for visit `n` to state `s` can be accessed as:

    ```TranspositionTable[s, n]```

    The full list of rewards for state `s` can be accessed as `TranspositionTable[s]`.
    """

    def __new__(cls, *args, **kwargs):
        # Idempotent
        if not kwargs and len(args) == 1 and isinstance(args[0], cls):
            return args[0]
        else:
            return super().__new__(cls)

    def __init__(
        self,
        table: Optional[DefaultMapping[str, list]] = None,
    ):
        if table is None:
            self._table = defaultdict(list)
        elif isinstance(table, DefaultMapping):
            self._table = table
        else:
            raise TypeError(
                f"table must be a DefaultMapping, not {type(table).__name__}"
            )

    @property
    def table(self):
        return self._table

    def __getitem__(self, state_wwo_visits):
        """Access the results of visit(s) to a state. If no visit number is specified,
        return the list of all visit results."""
        match state_wwo_visits:
            case state, slice(start=start, stop=stop, step=step):
                return self.table[state][start:stop:step]
            case state, [*visits]:
                state_table = self.table[state]
                return [state_table[visit] for visit in visits]
            case state, visit:
                return self.table[state][visit]
            case state:
                return self.table[state]

    def __missing__(self, state_wwo_visits):
        """If the state is not in the table, raise an error. If accessed without
        visit number(s), create an empty entry for this state."""
        n_visits = len(self.table[state])
        match state_wwo_visits:
            case state, slice(start=start, stop=stop, step=step):
                raise IndexError(
                    f"Cannot index slice({start}, {stop}, {step}) for state {state} "
                    f"with {n_visits} total visits."
                )
            case state, [*visits]:
                try:
                    for v in visits:
                        _ = self[state, v]
                except IndexError:
                    raise IndexError(
                        f"Visit index {v} does not exist for state {state} with "
                        f"{n_visits} total visits."
                    )
            case (state, visit):
                raise IndexError(
                    f"Visit index {visit} does not exist for state {state} with "
                    f"{n_visits} total visits."
                )
            case state:
                return self.table[state]  # defaultdict will create a new entry

    def __contains__(self, state_wwo_visit):
        n_visits = len(self.table[state])
        match state_wwo_visit:
            case state, slice(start=start, stop=stop, step=step):
                return state in self.table and stop <= n_visits
            case state, [*visits]:
                return state in self.table and all(v < n_visits for v in visits)
            case (state, visit):
                return state in self.table and visit < n_visits
            case state:
                return state in self.table

    def __len__(self):
        return len(self.table)

    def n_visits(self, state):
        """Return number of visits with triggering a default factory call."""
        if state in self.table:
            return len(self.table[state])
        else:
            return 0

    def keys(self):
        return self.table.keys()

    def values(self):
        return self.table.values()

    def items(self):
        return self.table.items()

    def draw_random_reward(
        self, state: str, rg: Optional[np.random.Generator] = None
    ) -> tuple[int, float]:
        """Draw a random visit number and resulting reward from state `state`."""
        if rg is None:
            rg = np.random.default_rng()
        index = rg.integers(0, len(self.table[state]))
        return index, self.table[state, index]

    def draw_bootstrap_reward(
        self, state: str, size: int, rg: Optional[np.random.Generator] = None
    ):
        """Draw a bootstrap sample of size `size` from the reward values for `state`."""
        if rg is None:
            rg = np.random.default_rng()
        indices = rg.integers(0, len(self.table[state]), size=size)
        state_table = self.table[state]
        return indices, np.array([state_table[i] for i in indices])

    @classmethod
    def _load_from_source(
        cls,
        read_func: Callable,
        src: Iterable[Path | str] | Path | str,
        progress: bool = False,
        state_col: str = "state",
        visit_col: str = "visit",
        reward_col: str = "reward",
        **load_kwargs,
    ):
        if isinstance(src, Iterable):
            if progress:
                from tqdm import tqdm

                iter_src = tqdm(src)
            else:
                iter_src = src
            df = pd.concat([read_func(f, **load_kwargs) for f in iter_src])
        else:
            df = read_func(src, **load_kwargs)

        if state_col in df.columns:
            df.set_index(state_col, inplace=True)

        if df.index.name == state_col:
            df.index = pd.CategoricalIndex(df.index, ordered=True)
        else:
            raise ValueError(
                f"Dataframe does not have a column or index named '{state_col}'."
            )

        if reward_col not in df.columns:
            raise ValueError(f"Dataframe does not have a column named '{reward_col}'.")

        if visit_col not in df.columns:
            df[visit_col] = df.groupby(state_col).cumcount()

        return cls.from_dataframe(
            df,
            visit_column=visit_col,
            reward_column=reward_col,
        )

    @classmethod
    def from_csv(cls, source, **kwargs):
        return cls._load_from_source(pd.read_csv, source, **kwargs)

    @classmethod
    def from_parquet(cls, source, **kwargs):
        return cls._load_from_source(pd.read_parquet, source, **kwargs)

    @classmethod
    def from_feather(cls, source, **kwargs):
        return cls._load_from_source(pd.read_feather, source, **kwargs)

    @classmethod
    def from_hdf(cls, source, **kwargs):
        return cls._load_from_source(pd.read_hdf, source, **kwargs)

    @classmethod
    def from_json(cls, source, **kwargs):
        return cls._load_from_source(pd.read_json, source, **kwargs)

    @classmethod
    def from_pickle(cls, source, **kwargs):
        return cls._load_from_source(pd.read_pickle, source, **kwargs)

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        visit_column: str = "visit",
        reward_column: str = "reward",
    ):
        table = defaultdict(list)
        grouped = (
            df[[visit_column, reward_column]]
            .sort_values(visit_column)
            .groupby(level=0)[reward_column]
        )
        for state, rewards in grouped:
            table[state] = rewards.to_list()
        return cls(table=table)

    def to_dataframe(self, state_col: str = "state", visit_col: str = "visit"):
        df = pd.concat(
            {k: pd.Series(v) for k, v in self.table.items()},
            names=[state_col, visit_col],
        ).reset_index()
        df[state_col] = df[state_col].astype("category")
        return df

    def to_csv(self, fname, **kwargs):
        self.to_dataframe().to_csv(fname, index=True, **kwargs)

    def to_parquet(self, fname, **kwargs):
        self.to_dataframe().to_parquet(fname, index=True, **kwargs)

    def to_hdf(self, fname, **kwargs):
        self.to_dataframe().to_hdf(fname, index=True, **kwargs)


class ParallelTree(CircuiTree):
    """A parallelizable implementation of CircuiTree. Parallelism is achieved
    via drawing multiple rewards at once. If boostrap is True, rewards for a
    set of visits to a state are drawn as bootstrap samples from the transposition
    table entries for that state. Otherwise, rewards are generated by indexing the
    transposition table by visit number. If desired results are not present in the table,
    they will be computed in parallel.

    Random seeds, parameter sets, and initial conditions are selected from the
    parameter table `self.param_table`.

    The methods `simulate_visits` and `save_results` must be implemented in a subclass.

    An invalid simulation result (e.g. a timeout) should be represented by a
    NaN reward value. If all rewards in a batch are NaNs, a new batch will be
    drawn from the transposition table. Otherwise, any NaN rewards will be
    ignored and the mean of the remaining rewards will be returned as the
    reward for the batch.
    """

    def __init__(
        self,
        *,
        parameter_table: Optional[ParameterTable] = None,
        transposition_table: Optional[TranspositionTable] = None,
        counter: Counter = None,
        seed_col: Optional[str] = None,
        param_cols: Optional[Iterable[str]] = None,
        init_cols: Optional[Iterable[str]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if isinstance(parameter_table, pd.DataFrame):
            self._parameter_table = ParameterTable.from_dataframe(
                parameter_table,
                seed_col=seed_col,
                param_cols=param_cols,
                init_cols=init_cols,
            )
        else:
            self._parameter_table = parameter_table
        self._transposition_table = TranspositionTable(transposition_table)

        self.visit_counter = Counter(counter)

        # Specify any attributes that should not be serialized when dumping to file
        self._non_serializable_attrs.extend(
            [
                "_parameter_table",
                "_transposition_table",
                "visit_counter",
            ]
        )

    @property
    def param_table(self) -> ParameterTable:
        return self._parameter_table

    @property
    def ttable(self) -> TranspositionTable:
        return self._transposition_table

    def reset_counter(self):
        self.visit_counter.clear()



from typing import Iterable
import numpy as np

__all__ = ["regret"]


def regret(outcomes: Iterable[int | float], optimal_payout: int | float = -1):
    if optimal_payout < 0:
        optimal_payout = np.max(outcomes)
    return np.cumsum(optimal_payout - outcomes)
